@@ -309,7 +309,7 @@ static int cmd_data_do_handle_input(struct smtp_server_cmd_ctx *cmd)
 		/* command is waiting for external event or it failed */
 		i_assert(smtp_server_command_is_replied(command));
 	}
-	
+
 	return 1;
 }
 
@@ -327,7 +327,7 @@ static int cmd_data_handle_input(struct smtp_server_cmd_ctx *cmd)
 
 	/* continue reading from client */
 	ret = cmd_data_do_handle_input(cmd);
-	
+
 	smtp_server_command_unref(&command);
 	smtp_server_connection_unref(&conn);
 
@@ -357,7 +357,7 @@ cmd_data_next(struct smtp_server_cmd_ctx *cmd,
 	/* DATA command stops the pipeline, so if it is next to reply, nothing
 	   else can be pending. */
 	i_assert(conn->state.pending_mail_cmds == 0 &&
-		conn->state.pending_rcpt_cmds == 0);
+		 conn->state.pending_rcpt_cmds == 0);
 
 	e_debug(cmd->event, "Command is next to be replied");
 
@@ -521,6 +521,7 @@ void smtp_server_cmd_data(struct smtp_server_cmd_ctx *cmd,
 	}
 
 	smtp_server_command_input_lock(cmd);
+	smtp_server_command_pipeline_block(cmd);
 
 	data_cmd = p_new(cmd->pool, struct cmd_data_context, 1);
 	data_cmd->chunk_first = TRUE;
@@ -580,6 +581,7 @@ int smtp_server_connection_data_chunk_add(struct smtp_server_cmd_ctx *cmd,
 	struct cmd_data_context *data_cmd = command->data;
 	uoff_t new_size;
 
+	i_assert(chunk != NULL);
 	i_assert(data_cmd != NULL);
 
 	if (trans != NULL)
@@ -604,6 +606,7 @@ int smtp_server_connection_data_chunk_add(struct smtp_server_cmd_ctx *cmd,
 		smtp_server_command_add_hook(
 			command, SMTP_SERVER_COMMAND_HOOK_REPLIED,
 			cmd_data_replied, data_cmd);
+		smtp_server_command_pipeline_block(cmd);
 	}
 
 	data_cmd->chunk_input = chunk;
