@@ -3,7 +3,7 @@
 #include "lib.h"
 #include "istream.h"
 #include "nfs-workarounds.h"
-#include "index-mail.h"
+#include "maildir-mail.h"
 #include "maildir-storage.h"
 #include "maildir-filename.h"
 #include "maildir-uidlist.h"
@@ -14,6 +14,8 @@
 #include <unistd.h>
 #include <sys/stat.h>
 
+static void maildir_mail_fix_corruption(struct mail *_mail);
+
 struct maildir_open_context {
 	int fd;
 	char *path;
@@ -31,7 +33,7 @@ do_open(struct maildir_mailbox *mbox, const char *path,
 	if (errno == ENOENT)
 		return 0;
 
-	if (errno == EACCES) {
+	if (ENOACCESS(errno)) {
 		mailbox_set_critical(&mbox->box, "%s",
 			mail_error_eacces_msg("open", path));
 	} else {
@@ -49,7 +51,7 @@ do_stat(struct maildir_mailbox *mbox, const char *path, struct stat *st)
 	if (errno == ENOENT)
 		return 0;
 
-	if (errno == EACCES) {
+	if (ENOACCESS(errno)) {
 		mailbox_set_critical(&mbox->box, "%s",
 			mail_error_eacces_msg("stat", path));
 	} else {
@@ -86,6 +88,7 @@ maildir_open_mail(struct maildir_mailbox *mbox, struct mail *mail,
 
 	if (ctx.fd == -1) {
 		*deleted_r = TRUE;
+		i_assert(ctx.path == NULL);
 		return NULL;
 	}
 
@@ -119,7 +122,8 @@ static int maildir_mail_stat(struct mail *mail, struct stat *st_r)
 		struct istream *input;
 
 		(void)mail_get_stream(mail, NULL, NULL, &input);
-	}
+	} else
+		mail_metadata_accessed_event(mail_event(mail));
 
 	if (imail->data.stream != NULL &&
 	    (fd = i_stream_get_fd(imail->data.stream)) != -1) {
@@ -223,6 +227,7 @@ static int maildir_get_pop3_state(struct index_mail *mail)
 	enum mail_cache_decision_type dec, vsize_dec;
 	enum mail_fetch_field allowed_pop3_fields;
 	bool not_pop3_only = FALSE;
+	pool_t pool;
 
 	if (mail->pop3_state_set)
 		return mail->pop3_state;
@@ -247,8 +252,7 @@ static int maildir_get_pop3_state(struct index_mail *mail)
 	} else {
 		/* also check if there are any non-[pv]size cached fields */
 		vsize_dec = MAIL_CACHE_DECISION_NO;
-		fields = mail_cache_register_get_list(box->cache,
-						      pool_datastack_create(),
+		fields = mail_cache_register_get_list(box->cache, &pool,
 						      &count);
 		for (i = 0; i < count; i++) {
 			dec = fields[i].decision & ENUM_NEGATE(MAIL_CACHE_DECISION_FORCED);
@@ -258,6 +262,7 @@ static int maildir_get_pop3_state(struct index_mail *mail)
 				 fields[i].idx != psize_idx)
 				not_pop3_only = TRUE;
 		}
+		pool_unref(&pool);
 	}
 
 	if (index_mail_get_vsize_extension(&mail->mail.mail) != NULL) {
@@ -591,7 +596,7 @@ maildir_mail_get_special(struct mail *_mail, enum mail_fetch_field field,
 		return index_mail_get_special(_mail, field, value_r);
 	}
 }
-							
+
 static int
 maildir_mail_get_stream(struct mail *_mail, bool get_body ATTR_UNUSED,
 			struct message_size *hdr_size,
@@ -603,6 +608,7 @@ maildir_mail_get_stream(struct mail *_mail, bool get_body ATTR_UNUSED,
 	struct index_mail_data *data = &mail->data;
 	bool deleted;
 
+	maildir_mail_fix_corruption(_mail);
 	if (data->stream == NULL) {
 		data->stream = maildir_open_mail(mbox, _mail, &deleted);
 		if (data->stream == NULL) {
@@ -758,20 +764,40 @@ maildir_mail_remove_sizes_from_filename(struct mail *mail,
 	(void)maildir_file_do(mbox, mail->uid, do_fix_size, &ctx);
 }
 
+static void maildir_mail_fix_corruption(struct mail *_mail)
+{
+	struct maildir_mail *mail =
+		container_of(_mail, struct maildir_mail, imail.mail.mail);
+
+	if (mail->corrupted_field == 0)
+		return;
+
+	mail->corrupted_field = 0;
+	maildir_mail_remove_sizes_from_uidlist(_mail);
+	maildir_mail_remove_sizes_from_filename(_mail, mail->corrupted_field);
+}
+
 static void maildir_mail_set_cache_corrupted(struct mail *_mail,
 					     enum mail_fetch_field field,
 					     const char *reason)
 {
+	struct maildir_mail *mail =
+		container_of(_mail, struct maildir_mail, imail.mail.mail);
+
 	if (field == MAIL_FETCH_PHYSICAL_SIZE ||
-	    field == MAIL_FETCH_VIRTUAL_SIZE) {
-		maildir_mail_remove_sizes_from_uidlist(_mail);
-		maildir_mail_remove_sizes_from_filename(_mail, field);
-	}
+	    field == MAIL_FETCH_VIRTUAL_SIZE)
+		mail->corrupted_field = field;
 	index_mail_set_cache_corrupted(_mail, field, reason);
 }
 
+static void maildir_mail_close(struct mail *_mail)
+{
+	maildir_mail_fix_corruption(_mail);
+	index_mail_close(_mail);
+}
+
 struct mail_vfuncs maildir_mail_vfuncs = {
-	index_mail_close,
+	maildir_mail_close,
 	index_mail_free,
 	index_mail_set_seq,
 	index_mail_set_uid,
